import json
import os
from datetime import timedelta

from airflow.contrib.operators.kubernetes_pod_operator import KubernetesPodOperator
from airflow.utils.dates import days_ago
from kubernetes.client import models as k8s

from airflow import DAG


def return_dag_ingrediants(project):
    """
    Args:
        project (str): The name of the project.

    Returns:
        paths (dict): A dictionary containing various file paths.
        default_args (dict): A dictionary containing default arguments for the DAG.
        envConfigMap (k8s.V1EnvFromSource): A reference to configmap that maps environment variables.
        volume (k8s.V1Volume): A Kubernetes volume.
        volume_mount (k8s.V1VolumeMount): A Kubernetes volume mount.
    """

    WORK_DIR = "/opt/aiola/projects"
    SUB_FOLDER = os.environ.get("DEPLOYMENT_DIR", "gad-deliveries")
    PROJECT_DIR = f"{WORK_DIR}/{SUB_FOLDER}/{project}"
    DBT_OUTPUT_DIR = "/opt/airflow/logs"
    PYTHON_DIR = f"{PROJECT_DIR}/python"
    DBT_DIR = f"{PROJECT_DIR}/dbt"
    CONFIG_DIR = f"{PROJECT_DIR}/configuration"

    paths = {
        "WORK_DIR": WORK_DIR,
        "SUB_FOLDER": SUB_FOLDER,
        "PROJECT_DIR": PROJECT_DIR,
        "DBT_DIR": DBT_DIR,
        "DBT_OUTPUT_DIR": DBT_OUTPUT_DIR,
        "PYTHON_DIR": PYTHON_DIR,
        "CONFIG_DIR": CONFIG_DIR,
    }

    default_args = {
        "owner": "GAD",
        "depends_on_past": False,
        "start_date": days_ago(0),
        "catchup": False,
        "retries": 0,
        "retry_delay": timedelta(seconds=10),
    }
    configMapEnvSource = k8s.V1ConfigMapEnvSource(
        name="gad-configmap",
        optional=False
    )
    envFromSource = k8s.V1EnvFromSource(
        config_map_ref=configMapEnvSource
    )
    # environment = []

    # environment = [
    #     k8s.V1EnvVar(name="PROJECTDIR", value="value1"),
    #     k8s.V1EnvVar(name="key2", value="value2"),
    # ]
    
    volume_mount = k8s.V1VolumeMount(
        name="project-volume",
        mount_path="/opt/aiola/projects",
        sub_path=None,
        read_only=True,
    )

    volume = k8s.V1Volume(
        name="project-volume",
        host_path=k8s.V1HostPathVolumeSource(path="/home/docker/projects"),
    )

    volumes = [volume]
    volumes_mounts = [volume_mount]

    return paths, default_args, envFromSource, volumes, volumes_mounts


def generate_airflow_dag(project: str, dag_id: str, schedule_interval, tasks: list):
    """
    Creates a DAG using the specified parameters.

    Args:
        project (str): The name of the project.
        dag_id (str): The ID of the DAG.
        schedule_interval (str): The schedule interval for the DAG.
        tasks (list): A list of dictionaries containing information about each task.

    Returns:
        dag (DAG): A DAG object.
    """

    paths, default_args, envConfigMap, volumes, volumes_mounts = return_dag_ingrediants(
        project
    )

    def return_image_name(task_type):
        """
        Returns the image name based on the task type.

        Parameters:
        task_type (str): A string representing the task type.

        Returns:
        str: A string representing the name of the image based on the task type.

        """
        if task_type == "dbt":
            return "gad-dbt:0.1"
        elif task_type == "python":
            return "gad-papermill:0.1"

    def return_cmds(task_dict: dict) -> list:
        """Returns a list of command-line commands based on task_dict.

        Args:
        task_dict: A dictionary containing information about the task to be executed.
                The dictionary must have 'task_type' key with value 'dbt' or 'python'.
                If 'task_type' is 'dbt', then the dictionary must have 'executable' key
                with a string value containing the name of the dbt executable to be run.
                If 'task_type' is 'python', then the dictionary must have 'executable' key
                with a string value containing the name of the python script to be run.

        Returns:
        A list of command-line commands based on the task type specified in task_dict.
        If task_type is 'dbt', then the returned list will contain ['dbt', <executable>]
        where <executable> is the value of 'executable' key in the task_dict.
        If task_type is 'python', then the returned list will contain ['python', <path/to/executable>]
        where <path/to/executable> is the full path to the python script specified in the
        'executable' key of the task_dict.
        """
        if task_dict["task_type"] == "dbt":
            return ["dbt", task_dict["executable"]]
        elif task_dict["task_type"] == "python":
            return ["python", f"{paths['PYTHON_DIR']}/{task_dict['executable']}.py"]

    def return_command_args(task_dict: dict, configs: dict) -> list:
        """Returns a list of command-line arguments based on task_dict and configs.

        Args:
        task_dict: A dictionary containing information about the task to be executed.
                The dictionary must have 'task_type' key with value 'dbt' or 'python'.
                If 'task_type' is 'dbt', then the dictionary must have 'dbt_models' key
                with a list of strings containing the names of dbt models to be executed.

        configs: A dictionary containing configuration values for the task.
                If the task is of type 'dbt', then the dictionary can have 'dbt_vars' key
                with a string value containing the variable values to be passed to dbt.
                If the task is of type 'python', then the dictionary can have 'python_args'
                key with a list of strings containing command-line arguments for the python script.

        Returns:
        A list of command-line arguments based on the task and configuration values.
        If task_type is 'dbt', then the returned list will contain arguments for dbt models
        and default dbt arguments such as project-dir, profiles-dir, target-path, and log-path.
        If task_type is 'python', then the returned list will contain arguments specified
        in the 'python_args' key of the configs dictionary.
        """

        if task_dict["task_type"] == "dbt":
            dbt_default_args = [
                "--project-dir",
                paths["DBT_DIR"],
                "--profiles-dir",
                paths["DBT_DIR"],
                "--target-path",
                paths["DBT_OUTPUT_DIR"],
                "--log-path",
                paths["DBT_OUTPUT_DIR"],
            ]

            dbt_default_args_and_models = task_dict["dbt_models"] + dbt_default_args

            dbt_vars = configs.get("dbt_vars")
            if dbt_vars:
                dbt_all_args = dbt_default_args_and_models + ["--vars", str(dbt_vars)]
            else:
                dbt_all_args = dbt_default_args_and_models

            return dbt_all_args

        elif task_dict["task_type"] == "python":
            python_args = configs.get("python_args")
            list_args = []
            if python_args:
                for key in python_args:
                    list_args.append(f"--{key}")
                    if python_args[key]:
                        list_args.append(python_args[key])
            return list_args

    def return_configs() -> dict:
        """
        Returns the dictionary of configurations read from a JSON file.

        Parameters:
        None

        Returns:
        dict: A dictionary containing the configurations read from the JSON file located in the CONFIG_DIR directory.

        """

        with open(f"{paths['CONFIG_DIR']}/config.json", "r") as f:
            j = f.read()
        return json.loads(j)

    # dag creation
    dag = DAG(
        dag_id=dag_id,
        default_args=default_args,
        schedule_interval=schedule_interval,
        max_active_runs=1,
        concurrency=10,
    )

    configs = return_configs()
    env_vars = configs.get("env_vars")

    # This code is a loop that iterates over a list of tasks and creates a KubernetesPodOperator object for each task.
    # return_command_args() function is used to obtain the command arguments for the task.
    # return_image_name() function is used to get the image name based on the task type.
    # return_configs() function is used to get environment variables.
    # The KubernetesPodOperator object is then created using these variables and appended to a dictionary named kubernetes_tasks with the task ID as the key.
    kubernetes_tasks = {}
    for task in tasks:
        cmds = return_cmds(task)
        arguments = return_command_args(task, configs)
        image = return_image_name(task["task_type"])

        kubernetes_task = KubernetesPodOperator(
            volumes=volumes,
            volume_mounts=volumes_mounts,
            env_vars=env_vars,
            env_from=[envConfigMap],
            namespace="default",
            labels={"Task": task["task_type"]},
            image_pull_policy="Never",
            name=task["task_id"],
            task_id=task["task_id"],
            is_delete_operator_pod=True,
            get_logs=True,
            image=image,
            cmds=cmds,
            arguments=arguments,
            dag=dag,
        )
        kubernetes_tasks[task["task_id"]] = kubernetes_task

    # using tge tasks list, and the kubernetes_tasks dictionary - this loop creates the dependancies.
    # each task in tasks contains a value in the 'upstream' key that tells what is the pervious task (or tasks).
    # the kubernates operator created gets the dependancies and is configured to use them with the set_upstream setting.
    for task in tasks:
        if task["upstream"] is None or task["upstream"] == "" or task["upstream"] == []:
            pass
        else:
            dependancies = []
            for t in task["upstream"]:
                dependancies.append(kubernetes_tasks[t])
            kubernetes_tasks[task["task_id"]].set_upstream(dependancies)

    return dag



import glob
import json
import logging
import os

import awswrangler as wr
import boto3
import pandas as pd
from botocore.exceptions import ClientError


class S3GeneralMethods:
    def split_s3_path(s3_path: str) -> str:
        """returns the bucket name and path

        Args:
            s3_path (str): s3 path

        Returns:
            str: bucket name
            str: s3 path (after the bucket name)
        """
        path_parts = s3_path.replace("s3://", "").split("/")
        bucket = path_parts.pop(0)
        key = "/".join(path_parts)
        return bucket, key


class S3Upload:
    def upload_dict_as_json_to_s3(dict_to_load: dict, s3_path: str) -> None:
        """load dict to s3 as json file

        Args:
            dict_to_load (dict): the dict to load
            s3_path (str): the path to load save the json file
        """
        df = pd.DataFrame.from_dict(data=dict_to_load.items(), orient="columns").T
        df.columns = df.iloc[0]
        df.drop(df.index[0], inplace=True)

        wr.s3.to_json(df=df, path=s3_path, index=False, orient="records", lines=True)
        logging.info(f"file was saved in s3: {s3_path}")

    def upload_s3_folder(
        s3_url: str, local_dir: str, file_extension: str = "*"
    ) -> None:
        """Uploads a local folder to an S3 bucket.

        Args:
            s3_url (str): The URL of the S3 bucket and folder.
            local_dir (str): The path of the local folder to be uploaded.
            file_extension (str, optional): The extension of the files to be uploaded. Defaults to '*'.

        Returns:
            None
        """
        s3 = boto3.resource("s3")
        bucket_name = S3GeneralMethods.split_s3_path(s3_path=s3_url)[0]

        # add '*/' so the iglob function will work as expected (in case there is no trailing slash in the given param)
        local_dir = os.path.join(local_dir, "*/")

        for file in glob.iglob(local_dir + f"**/*{file_extension}", recursive=True):
            file_name = str(file)
            s3.meta.client.upload_file(
                Filename=file_name, Bucket=bucket_name, Key=file_name
            )


class S3Read:
    def get_json_file_from_s3_as_dict(s3_path: str) -> dict:
        """retruns a dict containing the json file in the given s3 path.
        if no file is found an empty dict is returned

        Args:
            s3_path (str): full path to json file in s3 (e.g. "s3://bucket/path/to/file.json")

        Returns:
            dict: a df containing the json file
        """
        s3 = boto3.resource("s3")

        bucket_name, path = S3GeneralMethods.split_s3_path(s3_path)
        content_object = s3.Object(bucket_name=bucket_name, key=path)

        try:
            file_content = content_object.get()["Body"].read().decode("utf-8")
        except ClientError as ex:
            if ex.response["Error"]["Code"] == "NoSuchKey":
                logging.info("No object found - returning empty")
                return dict()
            else:
                raise

        return json.loads(file_content)

    def download_s3_folder(s3_url: str, dest_root_folder: str = None) -> None:
        """
        Downloads a folder from Amazon S3 to the local file system.

        Args:
            s3_url (str): The URL of the S3 folder, e.g. 's3://my-bucket/my-folder/'.
            dest_root_folder (str, optional): The local directory where the folder will be saved. If not specified, the folder will be saved in the current working directory.

        Returns:
            None
        """
        s3 = boto3.resource("s3")
        bucket_name, s3_folder = S3GeneralMethods.split_s3_path(s3_path=s3_url)
        bucket = s3.Bucket(bucket_name)

        for obj in bucket.objects.filter(Prefix=s3_folder):
            if dest_root_folder:
                # the target is the full s3 path of the file, but we replace the first folder in the
                # path with the dest_root_folder
                target = os.path.join(
                    dest_root_folder, os.path.relpath(obj.key, dest_root_folder)
                )
            else:
                # the target is the full s3 path of the file
                target = obj.key
            if not os.path.exists(os.path.dirname(target)):
                os.makedirs(os.path.dirname(target))
            if obj.key[-1] == "/":
                continue
            bucket.download_file(obj.key, target)



import json
import logging
from decimal import Decimal

import boto3
from boto3.dynamodb.conditions import Key


class DataTypesEncoder(json.JSONEncoder):
    def default(self, o: any) -> any:
        """convert decimal to string and set to list

        Args:
            o (any): the itme to encode

        Returns:
            any: the item after encoding
        """
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, set):
            return list(o)
        return super(DataTypesEncoder, self).default(o)


class DynamoDbTable:
    def __init__(self, table_name: str, region: str) -> None:
        """this constructor sets the following class attributes:
        - table_name
        - ddb_table
        - ddb_table_partition_key
        - ddb_table_sort_key
        - table_keys

        Args:
            table_name (str): name of the dynamodb table
            region (str): aws region of the table
        """
        dynamo_db_resource = boto3.resource("dynamodb", region_name=region)
        self.ddb_table = dynamo_db_resource.Table(table_name)
        self.table_name = table_name

        self.ddb_table_partition_key = self.ddb_table.key_schema[0]["AttributeName"]

        try:
            self.ddb_table_sort_key = self.ddb_table.key_schema[1]["AttributeName"]
        except:
            self.ddb_table_sort_key = None

        if self.ddb_table_sort_key:
            self.table_keys = {
                "partition_key_name": self.ddb_table_partition_key,
                "sort_key_name": self.ddb_table_sort_key,
            }
        else:
            self.table_keys = {"partition_key_name": self.ddb_table_partition_key}

    def get_first_line_from_ddb(self) -> dict:
        """get the first record from the given dynamodb table. If no lines in the table an empty dict is returned

        Args:
            ddb_tbl: dynamodb table object

        Returns:
            dict: first record of the given table after json serialization
        """
        response = self.ddb_table.scan(Limit=1)
        if response["Items"]:
            regular_json_from_ddb = response["Items"][0]
            # serialize the dict to be valid json (e.g. replace set to list)
            regular_json_from_ddb_str = json.dumps(
                regular_json_from_ddb, cls=DataTypesEncoder
            )
            return json.loads(regular_json_from_ddb_str)
        else:
            return {}

    def get_specific_line_from_ddb(ddb_tbl, keys: dict) -> dict:
        """get 1 specific record from the given dynamodb table and given keys

        Args:
            ddb_tbl: dynamodb table object
            keys (dict): partition_key and sort_key (if exists) name and value

        Returns:
            dict: 1 specifc record of the given table after json serialization
        """
        response = ddb_tbl.get_item(Key=keys)
        regular_json_from_ddb = response["Item"]
        # serialize the dict to be valid json (e.g. replace set to list)
        regular_json_from_ddb_str = json.dumps(
            regular_json_from_ddb, cls=DataTypesEncoder
        )
        return json.loads(regular_json_from_ddb_str)

    @staticmethod
    def generate_update_expression(attributes_to_update: list) -> str:
        """returns an update_expression that gives alias per attribute

        Args:
            attributes_to_update (dict): the keys are the attributes to update

        Returns:
            str: update_expression to update dynamodb table with alias per attribute
        """
        temp_list = []
        # we add # to keys to avoid using reserved keywords (https://stackoverflow.com/a/74616856/7162781)
        temp_list = [f"#{k}=:_{k}" for k in attributes_to_update]
        return "set " + ", ".join(temp_list)

    @staticmethod
    def generate_expression_attribute_values(attributes_to_update: dict) -> dict:
        """returns expression_attribute_values that sets value per attribute

        Args:
            attributes_to_update (dict): the keys are the attributes to update

        Returns:
            dict: expression_attribute_values dictionary:
                    the keys are the aliases of attributes to update
                    the values are the values to set
        """
        expression_attribute_values = {}
        for k, v in attributes_to_update.items():
            expression_attribute_values[f":_{k}"] = v

        return expression_attribute_values

    @staticmethod
    def generate_expression_attribute_names(attributes_to_update: dict) -> dict:
        """returns expression_attribute_names that replaces the name of the attribute

        Args:
            attributes_to_update (dict): the keys are the attributes to update

        Returns:
            dict: the keys are the replacement and the values are the original attribute name (e.g. {'#comment': 'comment'})
        """
        return {f"#{attr}": attr for attr in attributes_to_update}

    def update_dynamodb(self, ddb_record: dict, attributes_to_update: dict) -> None:
        """update the attributes_to_update of the given record in the given dynamodb table

        Args:
            ddb_record (dict): record to update
            attributes_to_update (dict): attributes to update
        """

        ddb_table_keys = {}
        ddb_table_keys[self.ddb_table_partition_key] = ddb_record[
            self.ddb_table_partition_key
        ]
        if self.ddb_table_sort_key:
            ddb_table_keys[self.ddb_table_sort_key] = ddb_record[
                self.ddb_table_sort_key
            ]
        update_expression = DynamoDbTable.generate_update_expression(
            attributes_to_update=list(attributes_to_update.keys())
        )
        expression_attribute_values = (
            DynamoDbTable.generate_expression_attribute_values(
                attributes_to_update=attributes_to_update
            )
        )
        expression_attribute_names = DynamoDbTable.generate_expression_attribute_names(
            attributes_to_update=attributes_to_update
        )

        # update the record in dynamodb
        response = self.ddb_table.update_item(
            Key=ddb_table_keys,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=expression_attribute_values,
            ExpressionAttributeNames=expression_attribute_names,
            ReturnValues="UPDATED_NEW",
        )

        logging.info(
            f"the following attributes were updated in the table {self.ddb_table}: {response['Attributes']}"
        )

    def get_all_table_rows(self) -> list:
        """get all rows of the given dynamodb table

        Args:
            table (dynamodb table object): table to get the rows from

        Returns:
            list: all table rows
        """
        response = self.ddb_table.scan()
        rows = response["Items"]
        while "LastEvaluatedKey" in response:
            response = self.ddb_table.scan(
                ExclusiveStartKey=response["LastEvaluatedKey"]
            )
            rows.extend(response["Items"])
        return rows

    def get_specific_row(
        self,
        partition_key_value: str,
        sort_key_value: str,
    ) -> dict:
        """get a specifc row of a dynamodb table by its keys

        Args:
            table (dynamodb table object): table to get the row from
            partition_key_col_name (str): partition key name of the table
            partition_key_value (str): partition key value of a specific row
            sort_key_col_name (str): sort key name of the table
            sort_key_value (str): sort key value of a specific row

        Returns:
            dict: 1 dynamodb record
        """
        return self.ddb_table.query(
            KeyConditionExpression=Key(self.ddb_table_partition_key).eq(
                partition_key_value
            )
            & Key(self.ddb_table_sort_key).eq(sort_key_value)
        )["Items"]



import os
from pathlib import Path


class filesMethods:
    def create_local_folder(file_path: str) -> None:
        """
        Create a local folder at the specified path.

        Args:
            file_path (str): Path to the folder to be created.

        Returns:
            None
        """
        folder_path = os.path.dirname(file_path)
        Path(folder_path).mkdir(parents=True, exist_ok=True)

    def list_files_with_ext_in_dir(base_dir, extensions_list):
        """
        Get a list of all files with the specified extensions in a directory.

        Args:
            base_dir (str): Base directory to search for files.
            extensions_list (list): List of file extensions to search for, without the '.'.

        Returns:
            list: List of file paths that match the specified extensions.
        """
        file_list = []
        for path in Path(base_dir).glob(r"**/*"):
            if path.suffix in extensions_list:
                file_list.append(str(path))

        file_list.sort()
        return file_list

    def get_file_extension_without_dot(file_name: str) -> str:
        """
        Get the file extension of a file, without the '.'.

        Args:
            file_name (str): Name of the file, can be either "name.ext" or "path/to/file/name.ext".

        Returns:
            str: The file extension without the '.'.
        """
        return os.path.splitext(file_name)[-1][1:]



import logging

import pandas as pd
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import Session


class runAthenaQuery:
    def __init__(
        self,
        schema_name: str,
        s3_staging_dir: str,
        region_name: str,
        work_group: str = "primary",
    ) -> None:
        """set up the connection to athena using sqlalchemy

        Args:
            schema_name (str): schema name
            s3_staging_dir (str): s3 pathto store the queryies output
            region_name (str): aws region name
            work_group (str, optional): the athena workgroup. Defaults to 'primary'.
        """

        # Create the SQLAlchemy connection. Note that you need to have pyathena installed for this.
        engine = create_engine(
            f"""awsathena+rest://:@athena.{region_name}.amazonaws.com:443/{schema_name}?s3_staging_dir{s3_staging_dir}&work_group={work_group}"""
        )

        self.session = Session(engine, future=True)
        self.conn = engine.connect()

    def run_query_with_pandas(self, query: str) -> pd.DataFrame:
        """execute query with pandas built in method: read_sql_query()

        Args:
            query (str): the query to be executed

        Returns:
            pd.DataFrame: dataframe containing the results of the query
        """
        return pd.read_sql_query(query, self.conn)

    def run_query(
        self, query: str, get_results: bool = True, print_query: bool = False
    ) -> list:
        """execute the query in athena

        Args:
            query (str): the query to be executed
            get_results (bool, optional): whether to return the output of the query. Defaults to True.
            print_query (bool, optional): True to print it, False not. Defaults to False.

        Returns:
            list: if get_results set to True then we return the output of the query. Else, nothing is returned
        """
        if print_query:
            logging.info(f"query to be executed: {query}")

        if get_results:
            return self.session.execute(query).scalars().all()
        else:
            self.session.execute(query)
            return
